@@ -339,6 +339,8 @@ udisks_linux_drive_ata_update (UDisksLinuxDriveAta    *drive,
   update_security (drive, device);
 
  out:
+  /* ensure property changes are sent before the method return */
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (drive));
   if (device != NULL)
     g_object_unref (device);
 
@@ -580,6 +582,7 @@ udisks_linux_drive_ata_refresh_smart_sync (UDisksLinuxDriveAta  *drive,
                        UDISKS_ERROR,
                        UDISKS_ERROR_FAILED,
                        "sk_disk_open: %m");
+          g_free (blob);
           goto out;
         }
 
@@ -681,6 +684,9 @@ udisks_linux_drive_ata_refresh_smart_sync (UDisksLinuxDriveAta  *drive,
   /* update stats again to account for the IO we just did to read the SMART info */
   update_io_stats (drive, device);
 
+  /* ensure property changes are sent before the method return */
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (drive));
+
  out:
   g_clear_object (&device);
   if (d != NULL)
@@ -1619,10 +1625,13 @@ apply_conf_data_free (ApplyConfData *data)
   g_free (data);
 }
 
-static gpointer
-apply_configuration_thread_func (gpointer user_data)
+static void
+apply_configuration_thread_func (GTask        *task,
+                                 gpointer      source_object,
+                                 gpointer      task_data,
+                                 GCancellable *cancellable)
 {
-  ApplyConfData *data = user_data;
+  ApplyConfData *data = task_data;
   UDisksDaemon *daemon;
   const gchar *device_file = NULL;
   gint fd = -1;
@@ -1793,8 +1802,6 @@ apply_configuration_thread_func (gpointer user_data)
  out:
   if (fd != -1)
     close (fd);
-  apply_conf_data_free (data);
-  return NULL;
 }
 
 /**
@@ -1813,6 +1820,7 @@ udisks_linux_drive_ata_apply_configuration (UDisksLinuxDriveAta *drive,
 {
   gboolean has_conf = FALSE;
   ApplyConfData *data = NULL;
+  GTask *task;
 
   data = g_new0 (ApplyConfData, 1);
   data->ata_pm_standby = -1;
@@ -1856,9 +1864,10 @@ udisks_linux_drive_ata_apply_configuration (UDisksLinuxDriveAta *drive,
   /* this can easily take a long time and thus block (the drive may be in standby mode
    * and needs to spin up) - so run it in a thread
    */
-  g_thread_new ("apply-conf-thread",
-                apply_configuration_thread_func,
-                data);
+  task = g_task_new (data->object, NULL, NULL, NULL);
+  g_task_set_task_data (task, data, (GDestroyNotify) apply_conf_data_free);
+  g_task_run_in_thread (task, apply_configuration_thread_func);
+  g_object_unref (task);
 
   data = NULL; /* don't free data below */
 
@@ -1905,7 +1914,7 @@ on_secure_erase_update_progress_timeout (gpointer user_data)
  * @enhanced: %TRUE to use the enhanced version of the ATA secure erase command.
  * @error: Return location for error or %NULL.
  *
- * Performs an ATA Secure Erase opeartion. Blocks the calling thread until the operation completes.
+ * Performs an ATA Secure Erase operation. Blocks the calling thread until the operation completes.
  *
  * This operation may take a very long time (hours) to complete.
  *
@@ -2177,7 +2186,7 @@ udisks_linux_drive_ata_secure_erase_sync (UDisksLinuxDriveAta  *drive,
                                          &output,
                                          &cleanup_error))
         {
-          udisks_critical ("Failed to clear user password '%s' on %s (%s) while attemping clean-up after a failed secure erase operation. You may need to manually unlock the drive. The error was: %s (%s, %d)",
+          udisks_critical ("Failed to clear user password '%s' on %s (%s) while attempting clean-up after a failed secure erase operation. You may need to manually unlock the drive. The error was: %s (%s, %d)",
                         pass,
                         device_file,
                         udisks_drive_get_id (_drive),
