@@ -38,6 +38,7 @@
 #include "udiskslinuxprovider.h"
 #include "udiskslinuxmdraidobject.h"
 #include "udiskslinuxmdraid.h"
+#include "udiskslinuxmdraidhelpers.h"
 #include "udiskslinuxblockobject.h"
 #include "udisksdaemon.h"
 #include "udisksstate.h"
@@ -128,68 +129,6 @@ udisks_linux_mdraid_new (void)
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-static gchar *
-read_sysfs_attr (GUdevDevice *device,
-                 const gchar *attr)
-{
-  gchar *ret = NULL;
-  gchar *path = NULL;
-  GError *error = NULL;
-
-  g_return_val_if_fail (G_UDEV_IS_DEVICE (device), NULL);
-
-  path = g_strdup_printf ("%s/%s", g_udev_device_get_sysfs_path (device), attr);
-  if (!g_file_get_contents (path, &ret, NULL /* size */, &error))
-    {
-      udisks_warning ("Error reading sysfs attr `%s': %s (%s, %d)",
-                      path, error->message, g_quark_to_string (error->domain), error->code);
-      g_clear_error (&error);
-      goto out;
-    }
-
- out:
-  g_free (path);
-  return ret;
-}
-
-static gint
-read_sysfs_attr_as_int (GUdevDevice *device,
-                        const gchar *attr)
-{
-  gint ret = 0;
-  gchar *str = NULL;
-
-  str = read_sysfs_attr (device, attr);
-  if (str == NULL)
-    goto out;
-
-  ret = atoi (str);
-  g_free (str);
-
- out:
-  return ret;
-}
-
-static guint64
-read_sysfs_attr_as_uint64 (GUdevDevice *device,
-                           const gchar *attr)
-{
-  guint64 ret = 0;
-  gchar *str = NULL;
-
-  str = read_sysfs_attr (device, attr);
-  if (str == NULL)
-    goto out;
-
-  ret = atoll (str);
-  g_free (str);
-
- out:
-  return ret;
-}
-
-/* ---------------------------------------------------------------------------------------------------- */
-
 static gboolean
 on_polling_timout (gpointer user_data)
 {
@@ -306,9 +245,9 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
   guint64 sync_remaining_time = 0;
   GVariantBuilder builder;
   UDisksDaemon *daemon = NULL;
-  gboolean has_redundancy = FALSE;
-  gboolean has_stripes = FALSE;
   UDisksBaseJob *job = NULL;
+  GError *error = NULL;
+  BDMDExamineData *raid_data = NULL;
 
   daemon = udisks_linux_mdraid_object_get_daemon (object);
 
@@ -330,33 +269,17 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
     {
       device = UDISKS_LINUX_DEVICE (member_devices->data);
       num_devices = g_udev_device_get_property_as_int (device->udev_device, "UDISKS_MD_MEMBER_DEVICES");
-      if (! num_devices)
-          num_devices = g_udev_device_get_property_as_int (device->udev_device, "STORAGED_MD_MEMBER_DEVICES");
       level = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_LEVEL");
-      if (! level)
-          level = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_LEVEL");
       uuid = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_UUID");
-      if (! uuid)
-          uuid = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_UUID");
       name = g_udev_device_get_property (device->udev_device, "UDISKS_MD_MEMBER_NAME");
-      if (! name)
-          name = g_udev_device_get_property (device->udev_device, "STORAGED_MD_MEMBER_NAME");
     }
   else
     {
       device = UDISKS_LINUX_DEVICE (raid_device);
       num_devices = g_udev_device_get_property_as_int (device->udev_device, "UDISKS_MD_DEVICES");
-      if (! num_devices)
-          num_devices = g_udev_device_get_property_as_int (device->udev_device, "STORAGED_MD_DEVICES");
       level = g_udev_device_get_property (device->udev_device, "UDISKS_MD_LEVEL");
-      if (! level)
-          level = g_udev_device_get_property (device->udev_device, "STORAGED_MD_LEVEL");
       uuid = g_udev_device_get_property (device->udev_device, "UDISKS_MD_UUID");
-      if (! uuid)
-          uuid = g_udev_device_get_property (device->udev_device, "STORAGED_MD_UUID");
       name = g_udev_device_get_property (device->udev_device, "UDISKS_MD_NAME");
-      if (! name)
-          name = g_udev_device_get_property (device->udev_device, "STORAGED_MD_NAME");
     }
 
   /* figure out size */
@@ -366,7 +289,15 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
     }
   else
     {
-      /* TODO: need MD_ARRAY_SIZE, see https://bugs.freedesktop.org/show_bug.cgi?id=53239#c5 */
+      raid_data = bd_md_examine (g_udev_device_get_device_file (device->udev_device),
+                                 &error);
+      if (raid_data == NULL)
+        {
+          udisks_debug ("Failed to read array size: %s", error->message);
+          g_clear_error (&error);
+        }
+      else
+          size = raid_data->size;
     }
 
   udisks_mdraid_set_uuid (iface, uuid);
@@ -377,38 +308,18 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
 
   udisks_mdraid_set_running (iface, raid_device != NULL);
 
-  if (g_strcmp0 (level, "raid1") == 0 ||
-      g_strcmp0 (level, "raid4") == 0 ||
-      g_strcmp0 (level, "raid5") == 0 ||
-      g_strcmp0 (level, "raid6") == 0 ||
-      g_strcmp0 (level, "raid10") == 0)
-    has_redundancy = TRUE;
-
-  if (g_strcmp0 (level, "raid0") == 0 ||
-      g_strcmp0 (level, "raid4") == 0 ||
-      g_strcmp0 (level, "raid5") == 0 ||
-      g_strcmp0 (level, "raid6") == 0 ||
-      g_strcmp0 (level, "raid10") == 0)
-    has_stripes = TRUE;
-
   if (raid_device != NULL)
     {
-      if (has_redundancy)
+      if (mdraid_has_redundancy (level))
         {
           /* Can't use GUdevDevice methods as they cache the result and these variables vary */
           degraded = read_sysfs_attr_as_int (raid_device->udev_device, "md/degraded");
           sync_action = read_sysfs_attr (raid_device->udev_device, "md/sync_action");
-          if (sync_action != NULL)
-            g_strstrip (sync_action);
           sync_completed = read_sysfs_attr (raid_device->udev_device, "md/sync_completed");
-          if (sync_completed != NULL)
-            g_strstrip (sync_completed);
           bitmap_location = read_sysfs_attr (raid_device->udev_device, "md/bitmap/location");
-          if (bitmap_location != NULL)
-            g_strstrip (bitmap_location);
         }
 
-      if (has_stripes)
+      if (mdraid_has_stripes (level))
         {
           chunk_size = read_sysfs_attr_as_uint64 (raid_device->udev_device, "md/chunk_size");
         }
@@ -438,44 +349,41 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
         }
     }
 
-  if (sync_action)
+  if (sync_action == NULL || g_strcmp0 (sync_action, "idle") == 0)
     {
-      if (g_strcmp0 (sync_action, "idle") != 0)
+      if (udisks_linux_mdraid_object_has_sync_job (object))
         {
-          if (! udisks_linux_mdraid_object_has_sync_job (object))
-            {
-              /* Launch a job */
-              job = udisks_daemon_launch_simple_job (daemon,
-                                                     UDISKS_OBJECT (object),
-                                                     sync_action_to_job_id (sync_action),
-                                                     0,
-                                                     NULL /* cancellable */);
+          /* Complete the job */
+          udisks_linux_mdraid_object_complete_sync_job (object,
+                                                        TRUE,
+                                                        "Finished");
+        }
+    }
+  else
+    {
+      if (! udisks_linux_mdraid_object_has_sync_job (object))
+        {
+          /* Launch a job */
+          job = udisks_daemon_launch_simple_job (daemon,
+                                                 UDISKS_OBJECT (object),
+                                                 sync_action_to_job_id (sync_action),
+                                                 0,
+                                                 NULL /* cancellable */);
 
-              /* Mark the job as not cancellable. It simply has to finish... */
-              udisks_job_set_cancelable (UDISKS_JOB (job), FALSE);
-              udisks_linux_mdraid_object_set_sync_job (object, job);
-            }
-          else
-            job = udisks_linux_mdraid_object_get_sync_job (object);
-
-          /* Update the job's interface */
-          udisks_job_set_progress (UDISKS_JOB (job), sync_completed_val);
-          udisks_job_set_progress_valid (UDISKS_JOB (job), TRUE);
-          udisks_job_set_rate (UDISKS_JOB (job), sync_rate);
-
-          udisks_job_set_expected_end_time (UDISKS_JOB (job),
-                                            g_get_real_time () + sync_remaining_time);
+          /* Mark the job as not cancellable. It simply has to finish... */
+          udisks_job_set_cancelable (UDISKS_JOB (job), FALSE);
+          udisks_linux_mdraid_object_set_sync_job (object, job);
         }
       else
-        {
-          if (udisks_linux_mdraid_object_has_sync_job (object))
-            {
-              /* Complete the job */
-              udisks_linux_mdraid_object_complete_sync_job (object,
-                                                            TRUE,
-                                                            "Finished");
-            }
-        }
+        job = udisks_linux_mdraid_object_get_sync_job (object);
+
+      /* Update the job's interface */
+      udisks_job_set_progress (UDISKS_JOB (job), sync_completed_val);
+      udisks_job_set_progress_valid (UDISKS_JOB (job), TRUE);
+      udisks_job_set_rate (UDISKS_JOB (job), sync_rate);
+
+      udisks_job_set_expected_end_time (UDISKS_JOB (job),
+                                        g_get_real_time () + sync_remaining_time);
     }
   udisks_mdraid_set_sync_completed (iface, sync_completed_val);
   udisks_mdraid_set_sync_rate (iface, sync_rate);
@@ -549,13 +457,11 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
               member_state = read_sysfs_attr (raid_device->udev_device, buf);
               if (member_state != NULL)
                 {
-                  g_strstrip (member_state);
                   member_state_elements = g_strsplit (member_state, ",", 0);
                 }
               else
                 {
-                  // g_variant_new() doesn't like NULL pointers for empty
-                  // arrays
+                  /* g_variant_new() doesn't like NULL pointers for empty arrays */
                   member_state_elements = g_new0 (gchar*, 1);
                 }
 
@@ -563,7 +469,6 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
               member_slot = read_sysfs_attr (raid_device->udev_device, buf);
               if (member_slot != NULL)
                 {
-                  g_strstrip (member_slot);
                   if (g_strcmp0 (member_slot, "none") != 0)
                     member_slot_as_int = atoi (member_slot);
                 }
@@ -594,12 +499,11 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
           /* ... and finally build (builder consumes each GVariant instance) */
           for (n = 0; n < p->len; n++)
             g_variant_builder_add_value (&builder, p->pdata[n]);
-          g_ptr_array_free (p, TRUE);
 
           g_dir_close (md_dir);
         }
       g_free (md_dir_name);
-
+      g_ptr_array_free (p, TRUE);
     }
   udisks_mdraid_set_active_devices (iface, g_variant_builder_end (&builder));
 
@@ -608,11 +512,15 @@ udisks_linux_mdraid_update (UDisksLinuxMDRaid       *mdraid,
                                                                                 uuid));
 
  out:
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (mdraid));
+  if (raid_data)
+      bd_md_examine_data_free (raid_data);
   g_free (sync_completed);
   g_free (sync_action);
   g_free (bitmap_location);
   g_list_free_full (member_devices, g_object_unref);
   g_clear_object (&raid_device);
+  g_clear_error (&error);
   return ret;
 }
 
@@ -663,7 +571,6 @@ handle_start (UDisksMDRaid           *_mdraid,
   const gchar *action_id;
   const gchar *message;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   GList *member_devices = NULL;
   gchar *raid_device_file = NULL;
@@ -693,8 +600,6 @@ handle_start (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -747,7 +652,7 @@ handle_start (UDisksMDRaid           *_mdraid,
 
   if (!bd_md_activate (NULL, NULL, udisks_mdraid_get_uuid (UDISKS_MDRAID (mdraid)), opt_start_degraded, NULL, &error))
     {
-      g_prefix_error (&error, "Error starting RAID array:");
+      g_prefix_error (&error, "Error starting RAID array: ");
       g_dbus_method_invocation_take_error (invocation, error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
       goto out;
@@ -760,12 +665,12 @@ handle_start (UDisksMDRaid           *_mdraid,
                                                      wait_for_md_block_object,
                                                      object,
                                                      NULL,
-                                                     10, /* timeout_seconds */
+                                                     UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                      &error);
   if (block_object == NULL)
     {
       g_prefix_error (&error,
-                      "Error waiting for MD block device after starting array");
+                      "Error waiting for MD block device after starting array: ");
       g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
@@ -835,7 +740,6 @@ udisks_linux_mdraid_stop (UDisksMDRaid           *_mdraid,
   UDisksLinuxMDRaidObject *object;
   uid_t started_by_uid;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   UDisksBaseJob *job = NULL;
   const gchar *device_file = NULL;
@@ -855,8 +759,6 @@ udisks_linux_mdraid_stop (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                error))
     {
       ret = FALSE;
@@ -924,7 +826,7 @@ udisks_linux_mdraid_stop (UDisksMDRaid           *_mdraid,
 
   if (!bd_md_deactivate (device_file, error))
     {
-      g_prefix_error (error, "Error stopping RAID array %s:", device_file);
+      g_prefix_error (error, "Error stopping RAID array '%s': ", device_file);
       g_dbus_method_invocation_take_error (invocation, *error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, (*error)->message);
       ret = FALSE;
@@ -1033,7 +935,6 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
   const gchar *message;
   uid_t started_by_uid;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   const gchar *device_file = NULL;
   const gchar *member_device_file = NULL;
@@ -1062,8 +963,6 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -1151,7 +1050,7 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
 
   if (!bd_md_remove (device_file, member_device_file, set_faulty, NULL, &error))
     {
-      g_prefix_error (&error, "Error removing %s from RAID array %s:", device_file, member_device_file);
+      g_prefix_error (&error, "Error removing '%s' from RAID array '%s': ", device_file, member_device_file);
       g_dbus_method_invocation_take_error (invocation, error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
       goto out;
@@ -1164,7 +1063,7 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
       if (!bd_fs_wipe (member_device_file, TRUE, &error))
         {
           g_prefix_error (&error,
-                          "Error wiping  %s after removal from RAID array %s:",
+                          "Error wiping '%s' after removal from RAID array '%s': ",
                           member_device_file,
                           device_file);
           g_dbus_method_invocation_take_error (invocation, error);
@@ -1175,7 +1074,7 @@ handle_remove_device (UDisksMDRaid           *_mdraid,
   udisks_mdraid_complete_remove_device (_mdraid, invocation);
 
  out:
-  g_free (member_states);
+  g_strfreev (member_states);
   g_clear_object (&member_device_object);
   g_clear_object (&member_device);
   g_clear_object (&raid_device);
@@ -1199,7 +1098,6 @@ handle_add_device (UDisksMDRaid           *_mdraid,
   const gchar *message;
   uid_t started_by_uid;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   const gchar *device_file = NULL;
   const gchar *new_member_device_file = NULL;
@@ -1223,8 +1121,6 @@ handle_add_device (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -1300,7 +1196,7 @@ handle_add_device (UDisksMDRaid           *_mdraid,
 
   if (!bd_md_add (device_file, new_member_device_file, 0, NULL, &error))
     {
-      g_prefix_error (&error, "Error adding %s to RAID array %s:", device_file, new_member_device_file);
+      g_prefix_error (&error, "Error adding '%s' to RAID array '%s': ", new_member_device_file, device_file);
       g_dbus_method_invocation_take_error (invocation, error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
       goto out;
@@ -1333,7 +1229,6 @@ handle_set_bitmap_location (UDisksMDRaid           *_mdraid,
   const gchar *message;
   uid_t started_by_uid;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   const gchar *device_file = NULL;
   GError *error = NULL;
@@ -1354,8 +1249,6 @@ handle_set_bitmap_location (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -1421,7 +1314,7 @@ handle_set_bitmap_location (UDisksMDRaid           *_mdraid,
 
   if (!bd_md_set_bitmap_location (device_file, value, &error))
     {
-      g_prefix_error (&error, "Error setting bitmap on RAID array %s: ", device_file);
+      g_prefix_error (&error, "Error setting bitmap on RAID array '%s': ", device_file);
       g_dbus_method_invocation_take_error (invocation, error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
       goto out;
@@ -1453,7 +1346,6 @@ handle_request_sync_action (UDisksMDRaid           *_mdraid,
   const gchar *message;
   uid_t started_by_uid;
   uid_t caller_uid;
-  gid_t caller_gid;
   UDisksLinuxDevice *raid_device = NULL;
   GError *error = NULL;
   const gchar *device_file = NULL;
@@ -1474,8 +1366,6 @@ handle_request_sync_action (UDisksMDRaid           *_mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -1549,6 +1439,7 @@ handle_request_sync_action (UDisksMDRaid           *_mdraid,
       goto out;
     }
 
+  udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
   udisks_mdraid_complete_request_sync_action (_mdraid, invocation);
 
  out:
@@ -1568,7 +1459,6 @@ udisks_linux_mdraid_delete (UDisksMDRaid           *mdraid,
   UDisksLinuxMDRaidObject *object;
   UDisksDaemon *daemon;
   uid_t caller_uid;
-  gid_t caller_gid;
   const gchar *message;
   const gchar *action_id;
   gboolean teardown_flag = FALSE;
@@ -1595,8 +1485,6 @@ udisks_linux_mdraid_delete (UDisksMDRaid           *mdraid,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                error))
     {
       ret = FALSE;
@@ -1687,8 +1575,7 @@ udisks_linux_mdraid_delete (UDisksMDRaid           *mdraid,
 
       if (!bd_md_destroy (device, error))
         {
-          g_prefix_error (error, "Error wiping device %s:", device);
-          g_dbus_method_invocation_take_error (invocation, *error);
+          g_prefix_error (error, "Error wiping device '%s': ", device);
           ret = FALSE;
           goto out;
         }
@@ -1699,6 +1586,7 @@ udisks_linux_mdraid_delete (UDisksMDRaid           *mdraid,
  out:
   g_list_free_full (member_devices, g_object_unref);
   g_clear_object (&raid_device);
+  g_clear_object (&object);
   return ret;
 }
 
